import os
import re
import tempfile
import subprocess as sub

# Squelch warnings that the transformers package gives about not having PyTorch etc. installed
# We are importing it only for its tokenizers
os.environ["TRANSFORMERS_VERBOSITY"] = "error"
from transformers import LlamaTokenizer, AutoTokenizer

from ..utils import echo_err

# Note: This eventually could be replaced with Python bindings for llama.cpp:
# https://github.com/abetlen/llama-cpp-python

class LlamaCpp(object):
    HOME_DIR = os.path.expanduser('~')
    LLAMA_CPP_DIR = f"{HOME_DIR}/src/llama.cpp"
    LLAMA_CPP_THREADS = 16
    LLAMA_CPP_TMPDIR = f"{HOME_DIR}/tmp/llama_input"


    def __init__(self, config, opts, llama_cpp_bin='main', vocab_path=None, hf_model_id=None,
            tokenizer_model_path=None, alt_bos_token=None):
        self.llama_cpp_bin = llama_cpp_bin
        self.opts = opts
        self.llama_cpp_dir = config.get("LLAMA_CPP_DIR", self.LLAMA_CPP_DIR)
        self.cwd = self.llama_cpp_dir
        self.threads = config.get("LLAMA_CPP_THREADS", self.LLAMA_CPP_THREADS)
        self.tmp_dir = config.get("LLAMA_CPP_TMPDIR", self.LLAMA_CPP_TMPDIR)
        self.alt_bos_token = alt_bos_token

        self.tokenizer = None
        if vocab_path is not None:
            vocab_path = os.path.join(self.llama_cpp_dir, vocab_path)
            self.tokenizer = LlamaTokenizer(vocab_path, legacy = False)
        elif tokenizer_model_path is not None:
            tokenizer_model_path = os.path.join(self.llama_cpp_dir, tokenizer_model_path)
            self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_model_path)
        elif hf_model_id is not None:
            self.tokenizer = AutoTokenizer.from_pretrained(hf_model_id)


    def autotrim_prompt(self, input_text, prompt, prompt_max_tokens):
        """Creates a prompt and ensures it's within the token limit.

        `input_text` is substituted into `prompt` using str.format, at the
        {input} marker. 

        `prompt_max_tokens` is the token limit, usually $context_window - 4.
        
        Returns the full prompt with trimmed `input_text` already inserted, 
        or None if the process fails."""
        trimmed_text = input_text
        if self.tokenizer is None:
            echo_err("Neither `vocab_path` nor `hf_model_id` was given; therefore a tokenizer "
                "could not be loaded for this model.")
            return None
        while True:
            full_prompt = prompt.format(input = trimmed_text)
            if len(self.tokenizer.encode(full_prompt)) <= prompt_max_tokens: break
            # If it's too long, cleave off the last sentence-like piece, and try again
            sentences_and_periods = re.split(r'([.]\s+|[.]$)', trimmed_text)
            if len(sentences_and_periods) < 4:
                return None
            else: trimmed_text = ''.join(sentences_and_periods[:-4]) + '.'
        return full_prompt


    def _llama_error(self, err_msg):
        echo_err("\n".join(filter(lambda line: "error:" in line, err_msg.split("\n"))))
        return None


    def __call__(self, full_prompt):
        llama_bin = os.path.join(self.llama_cpp_dir, self.llama_cpp_bin)
        raw_llm_output = None
        os.makedirs(self.tmp_dir, exist_ok=True)

        # <s> is a special beginning-of-string (BOS) token. We don't allow it anywhere
        # except the beginning of the prompt.
        full_prompt = re.sub(r'(?<!^)<s>', '', full_prompt)
        # We want to return only the LLM-created output (and drop the prompt which is also 
        # echoed to STDOUT). If the prompt starts with <s>, this is treated as a BOS token and
        # not echoed. Otherwise, llama.cpp echoes one space " ", for unclear reasons.
        prompt_echo_length = len(full_prompt) + 1
        if full_prompt.startswith('<s>'): prompt_echo_length -= 4
        # Some models have a special BOS token (like Llama 3) that llama.cpp adds automatically
        # to the prompt. This is also echoed to STDOUT and must be trimmed.
        if self.alt_bos_token is not None: prompt_echo_length += len(self.alt_bos_token) - 2

        with tempfile.NamedTemporaryFile(dir=self.tmp_dir) as tmp_input:
            tmp_input.write(full_prompt.encode())
            tmp_input.flush()
            required_opts = f"-t {str(self.threads)}"
            cmd = f"{llama_bin} {required_opts} {self.opts} -f {tmp_input.name}"
            res = sub.run(cmd, stdout=sub.PIPE, stderr=sub.PIPE, shell=True, cwd=self.cwd)
            if res.returncode != 0: return self._llama_error(res.stderr.decode('utf-8'))
            raw_llm_output = res.stdout.decode('utf-8')[prompt_echo_length: ]

        return raw_llm_output.strip(" .\n")


import numpy as np
import math
import re
from click import echo

from .config import Config
from .db import LlacieDatabase
from .edw import EDWDatabase

from .strategies import find_strategies
from .tasks.abstract import TaskOutputType as out_t
from .evaluate import ConfusionMatrix
from .utils import echo_info, echo_warn, echo_err, echo_strategies, only_one_strategy

class App(object):
    def __init__(self):
        self.config = Config()
        self._db = None
        self._edw = None
        self.is_worker = False  

    @property
    def db(self):
        # Lazily initializes the connection to the database when needed
        if self._db is None:
            self._db = LlacieDatabase(self.config)
        return self._db
    
    @property
    def edw(self):
        # Lazily initializes the connection to the EDW database when needed
        if self._edw is None:
            self._edw = EDWDatabase(self.config)
        return self._edw

    @property
    def is_worker(self): 
        """Whether or not this is a worker instance of the App. Worker instances are created
        within bsub or slurm jobs by a master instance of the App."""
        return self._is_worker

    @is_worker.setter
    def is_worker(self, value):
        self._is_worker = bool(value)
        # Syncs with a variable in the config as well -- this way strategies can see it
        self.config['LLACIE_WORKER'] = self._is_worker


    # Allows this class to support the context manager protocol and be used
    # in `with` blocks to automatically clean up its own resources
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if self._db is not None:
            self._db.close()
        if self._edw is not None:
            self._edw.close()
    

    def _rebuild_note_metadata(self, overwrite=False):
        self.db.warn_or_truncate('notes', overwrite)

        echo("Getting metadata for all episodes of interest.")
        eps_df = self.db.get_all_episodes()
        admit_pt_ec_ID_to_ep_ID = dict(
            zip(eps_df['admitPatientEncounterID'], eps_df['FK_episode_id'])
        )
        dc_pt_ec_ID_to_ep_ID = dict(
            zip(eps_df['dcPatientEncounterID'], eps_df['FK_episode_id'])
        )
        
        echo("Importing metadata for all admission, consult, and progress notes from "
            "ADMISSION encounters for each matching episode in EDW.")
        for df in self.edw.gen_admission_enc_note_metadata(eps_df['admitPatientEncounterID']):
            df['FK_episode_id'] = [
                admit_pt_ec_ID_to_ep_ID[ec_id] for ec_id in df['PatientEncounterID']
            ]
            self.db.append_df_to_table(df, 'notes')

        echo("Importing metadata for all discharge notes from DISCHARGE encounters "
            "for each matching episode in EDW.")
        for df in self.edw.gen_discharge_enc_note_metadata(eps_df['dcPatientEncounterID']):
            df['FK_episode_id'] = [
                dc_pt_ec_ID_to_ep_ID[ec_id] for ec_id in df['PatientEncounterID']
            ]
            self.db.append_df_to_table(df, 'notes')

    ##################################################################
    # High-level tasks that may be exposed as CLI commands
    ##################################################################

    def import_notes_text(self, input_path, overwrite=False):
        echo_info(f"Info: Importing notes from text file {input_path}.")
        self.db.warn_or_truncate('notes', overwrite)

        with open(input_path, 'r') as f:
            content = f.read()

        notes = re.split(r'\n#{10,}\n', content)
        notes = [note.strip() for note in notes if note.strip()]
        echo_info(f"Found {len(notes)} notes to import.")

        if len(notes) == 0:
            echo_warn("Warn: No notes found in the file.")
            return

        self.db.import_notes(notes)
        echo_info(f"Info: Successfully imported {len(notes)} notes.")
    

    def import_notes_edw(self, overwrite=False, resume_note_text=False):
        if not resume_note_text:
            self._rebuild_note_metadata(overwrite)

        echo_info("Info: Updating the notes table with full note text from EDW.")
        notes_df = self.db.get_notes_without_text()
        for df in self.edw.gen_note_text(notes_df['NoteID']):
            self.db.update_table_from_df(df, 'notes', 'NoteID')


    ##########################
    ####### SECTIONS #########
    ##########################

    def list_sections_with_strategies(self, section_name=None, strategy_name=None):
        strategies = find_strategies(output_type=out_t.SECTION, task=section_name,
            name_glob=strategy_name)
        if len(strategies) == 0:
            return echo_err("Error: No strategies matching those parameters were found.")
        
        echo_strategies(strategies, title="Sections")


    def run_section_strategy(self, section_name=None, strategy_name=None, dry_run=False):
        strategies = find_strategies(output_type=out_t.SECTION, task=section_name, 
            name_glob=strategy_name)
        if not only_one_strategy(strategies, "Sections"): return

        strat = strategies[0](self.db, self.config)
        ids = strat.get_unfinished_ids()
        echo_info(f"Info: {len(ids)} notes will undergo section extraction")
        if dry_run: echo_info("Info: Dry run, exiting")
        else: strat.run(ids)
    

    ##########################
    ####### FEATURES #########
    ##########################

    def list_features_with_strategies(self, feature_name=None, strategy_name=None):
        strategies = find_strategies(output_type=out_t.FEATURE, task=feature_name,
            name_glob=strategy_name)
        if len(strategies) == 0:
            return echo_err("Error: No strategies matching those parameters were found.")

        echo_strategies(strategies, title="Features")


    def run_feature_strategy(self, feature_name=None, strategy_name=None, dry_run=False,
            annotated_for_task=None, max_note_ids=None, redo_older_than=None, note_ids=tuple()):
        strategies = find_strategies(output_type=out_t.FEATURE, task=feature_name, 
            name_glob=strategy_name)
        if not only_one_strategy(strategies, "Features"): return

        strat = strategies[0](self.db, self.config)
        # FIXME: Should have strategies check whether prereqs are finished.

        note_ids = np.array(note_ids)
        if len(note_ids) == 0:
            note_ids = strat.get_unfinished_ids(newer_than=redo_older_than)

        if not self.is_worker:
            if annotated_for_task is not None:
                annot_strategies = find_strategies(task=annotated_for_task)
                if len(annot_strategies) == 0:
                    return echo_err(f"No task exists with that name: {annotated_for_task}")
                annot_task = annot_strategies[0].task
                annot_note_ids = self.db.get_ids_annotated_for_task(annot_task, strat.task)
                note_ids = np.intersect1d(note_ids, annot_note_ids)

            echo_info(f"Info: {len(note_ids)} notes will undergo feature extraction")

        strat.run(note_ids, max_note_ids=max_note_ids, dry_run=dry_run)


    ################################
    ####### EPISODE LABELS #########
    ################################

    def list_episode_labels_with_strategies(self, episode_label_task=None, strategy_name=None):
        strategies = find_strategies(output_type=out_t.EPISODE_LABEL, task=episode_label_task,
            name_glob=strategy_name)
        if len(strategies) == 0:
            return echo_err("Error: No strategies matching those parameters were found.")

        echo_strategies(strategies, title="Episode labels")


    def run_episode_label_strategy(self, episode_label_task=None, strategy_name=None, 
            dry_run=False, annotated_for_task=None):
        strategies = find_strategies(output_type=out_t.EPISODE_LABEL, task=episode_label_task, 
            name_glob=strategy_name)
        if not only_one_strategy(strategies, "Episode labels"): return

        strat = strategies[0](self.db, self.config)
        # FIXME: Should have strategies check whether prereqs are finished.
        ep_ids = strat.get_unfinished_ids()

        if annotated_for_task is not None:
            annot_strategies = find_strategies(task=annotated_for_task)
            if len(annot_strategies) == 0:
                return echo_err(f"No task exists with that name: {annotated_for_task}")
            annot_task = annot_strategies[0].task
            annot_ep_ids = self.db.get_ids_annotated_for_task(annot_task, strat.task)
            ep_ids = np.intersect1d(ep_ids, annot_ep_ids)

        echo_info(f"Info: {len(ep_ids)} episodes will undergo label creation")
        if dry_run: echo_info("Info: Dry run, exiting")
        else: strat.run(ep_ids)


    def _echo_confusion_matrices(self, df_human, df_pred, vocab, other_human=None, n_resamples=None):
        label = f"of {other_human}'s" if other_human is not None else "LLM"
        echo(f"=============== Using top 10 {label} responses ================")
        conf_mat = ConfusionMatrix.from_episode_labels(df_human, df_pred, vocab, 
            max_line_num=10, other_human=other_human, n_resamples=n_resamples)
        echo(conf_mat)
        echo(f"=============== Using all {label} responses ================")
        conf_mat = ConfusionMatrix.from_episode_labels(df_human, df_pred, vocab, 
            max_line_num=math.inf, other_human=other_human, n_resamples=n_resamples)
        echo(conf_mat)


    def evaluate_episode_label_strategies(self, episode_label_task=None, strategy_name=None,
            human_username=None, other_human=None, also_labeled_by=None, bootstrap_samples=None):
        strategies = find_strategies(output_type=out_t.EPISODE_LABEL, task=episode_label_task,
            name_glob=strategy_name)
        if len(strategies) == 0:
            return echo_err("Error: No strategies matching those parameters were found.")

        human_annotator = human_username if human_username is not None else True
        seen_tasks = set()

        for strategy_cls in strategies:
            strategy = strategy_cls(self.db, self.config)
            vocab = strategy.task.vocab

            ep_ids = None
            if also_labeled_by is not None:
                df_also_labeled = self.db.get_episode_labels(strategy, human_annotated=also_labeled_by)
                ep_ids = df_also_labeled["FK_episode_id"].unique()

            df_human = self.db.get_episode_labels(strategy, ep_ids, human_annotated=human_annotator)
            ep_ids = df_human["FK_episode_id"].unique()

            if other_human is not None:
                if strategy.task.name in seen_tasks: continue
                seen_tasks.add(strategy.task.name)
                echo_info(f"Task {strategy.task.name}, {len(ep_ids)} episodes, vocab size {len(vocab)}")

                df_pred = self.db.get_episode_labels(strategy, ep_ids, human_annotated=other_human)
                if len(df_pred) == 0:
                    echo_warn(f"Warn: No 2nd human annotations for {strategy.task.name}, skipping")
                    continue
                # When evaluating inter-labeler reliability, don't penalize for any episodes not
                #   labeled by the other human
                other_human_ep_ids = df_pred["FK_episode_id"].unique()
                df_human = df_human[df_human["FK_episode_id"].isin(other_human_ep_ids)]
            else:
                echo_info(f"Strategy {strategy.name}, {len(ep_ids)} episodes, vocab size {len(vocab)}")
                
                df_pred = self.db.get_episode_labels(strategy, ep_ids, human_annotated=False)
                if len(df_pred) == 0:
                    echo_warn(f"Warn: No episode labels created for {strategy.name}, skipping")
                    continue

            if len(df_human) == 0:
                echo_warn(f"Warn: No human annotations for {strategy.task.name}, skipping")
                continue

            self._echo_confusion_matrices(df_human, df_pred, vocab, other_human, bootstrap_samples)


import sys
import json
import re
from os import path
from textwrap import dedent
from socket import gethostname

from .abstract import AbstractStrategy
from ..batch.slurm import SlurmJobManager
from ..inference.llama_cpp import LlamaCpp
from ..utils import PACKAGE_DIR, echo_warn, echo_info, strtobool

class AbstractLlamaCppStrategy(AbstractStrategy):
    """
    Abstract class for running llama.cpp with constant command-line parameters and with a 
    prespecified prompt template, across all of the specified notes, substituting one section
    for each note into the prompt template.

    The output is currently constrained to be a JSON array of strings.
    
    Supports the use of Slurm to schedule jobs across a cluster when SLURM_ENABLED = True.

    This class is intended to be subclassed and configured primarily by overriding the class 
    constants, and beyond that, methods could be overridden as well.
    """
    # can be, e.g., ShortHPISectionRegexStrategy from .section.hpi_short.regex
    SECTION_STRATEGY_CLASS = None

    SLURM_ENABLED = True

    # These are example default options based on the Llama 3 8B strategy for presenting symptoms
    LLAMA_CPP_BIN = 'main-e849648'            # which llama.cpp `main` executable to use
    LLAMA_CPP_THREADS = 16
    SLURM_THREADS_PER_JOB = 16
    LLAMA_MODEL_PATH = 'models/llama3/Meta-Llama-3-8B-Instruct-bf16.gguf'
    GRAMMAR_FILE = path.join(PACKAGE_DIR, 'grammars/json_arr_of_strings.gbnf')
    LLM_MAX_TOKENS = 2044
    LLAMA_CPP_OPTS = (f"-m {LLAMA_MODEL_PATH} -n 256 -c {LLM_MAX_TOKENS + 4} "
        f"--grammar-file {GRAMMAR_FILE} -r '<|eot_id|>'")
    TOKENIZER_MODEL_PATH = 'models/llama3'    # For tokenized input length estimation only.
    TOKENIZER_HF_MODEL_ID = None              # Either set this or TOKENIZER_MODEL_PATH.
    LLAMA_VOCAB_PATH = None                   # Should be None for everything except Llama 1
    NICE_MODEL_NAME = "Llama3 8B Instruct (bf16)"
    ALT_BOS_TOKEN = '<|begin_of_text|>'
    TRIM_EOT_TOKEN_REGEX = r'<[|]eot_id[|]>$' # If not None, used to trim end of LLM output

    LLM_PROMPT = dedent("""\
        <|start_header_id|>system<|end_header_id|>
        
        You are a helpful assistant that reads medical charts and answers questions about them. 
        Use only the information in the text provided to answer the question.
        If a patient denies something, do not include it in your answer.
        After you provide an answer, you immediately stop talking.
        <|eot_id|>
        <|start_header_id|>user<|end_header_id|>
        
        Read the following patient history and list the patient's presenting symptoms.
        Include only symptoms present now or within the past few days.
        Ignore any symptoms from past medical history or prior healthcare encounters.
        Give your answer as a JSON array containing up to ten strings.
        Each string contains between one and three words.

        {input}
        <|eot_id|>
        <|start_header_id|>assistant<|end_header_id|>
        
        """)


    def __init__(self, db, config, **options):
        super().__init__(db, config, **options)
        # Ensure this config option exists because LlamaCpp() needs it
        self.config['LLAMA_CPP_THREADS'] = config.get("LLAMA_CPP_THREADS", self.LLAMA_CPP_THREADS)
        # Ensure this config option exists because SlurmJobManager() uses it
        self.config['SLURM_THREADS_PER_JOB'] = config.get("SLURM_THREADS_PER_JOB",
            self.SLURM_THREADS_PER_JOB)
        
        if not isinstance(self.SECTION_STRATEGY_CLASS, type):
            raise NotImplementedError("You must specify a SECTION_STRATEGY_CLASS")

        self.sec_strat = self.SECTION_STRATEGY_CLASS(db, config, **options)
        self.sec_task_name = self.sec_strat.task.name
        self.slurm_enabled = strtobool(config.get("SLURM_ENABLED", self.SLURM_ENABLED))


    def _run_llm(self, note_ids):
        llm = LlamaCpp(
            self.config,
            self.LLAMA_CPP_OPTS,
            self.LLAMA_CPP_BIN, 
            vocab_path = self.LLAMA_VOCAB_PATH,
            hf_model_id = self.TOKENIZER_HF_MODEL_ID,
            tokenizer_model_path = self.TOKENIZER_MODEL_PATH,
            alt_bos_token = self.ALT_BOS_TOKEN
        )

        for note_id in note_ids:
            # Get the note section to work on
            row = self.db.get_note_section(note_id, self.sec_strat)
            if row is None: 
                raise RuntimeError(f"No {self.sec_task_name} section found for note id {note_id}")
            section_value = row.section_value
            self.start_timer()

            # Ensure the prompt will fit in the context window
            full_prompt = llm.autotrim_prompt(section_value, self.LLM_PROMPT, self.LLM_MAX_TOKENS)
            if full_prompt is None:
                echo_warn(f"Autotrimming failed for note id {note_id}")
                # Couldn't autotrim; likely a malformed section, but try using the whole thing
                full_prompt = self.LLM_PROMPT.format(input = section_value)
            
            # Run the prompt through llama.cpp
            echo_info(f"Running note id {note_id} thru the {self.NICE_MODEL_NAME} model")
            llm_output_raw = llm(full_prompt)
            if self.TRIM_EOT_TOKEN_REGEX is not None:
                llm_output_raw = re.sub(self.TRIM_EOT_TOKEN_REGEX, "", llm_output_raw)

            if llm_output_raw is not None and len(llm_output_raw) > 0: 
                # Cleanup and parse the output, and save everything back to postgres.
                # Since we constrained output to be a JSON array of strings, parsing is simple
                try:
                    parsed = json.loads(llm_output_raw)
                    feature = "\n".join(parsed)
                except (json.decoder.JSONDecodeError, TypeError):
                    echo_warn(f"Could not decode LLM output as JSON for note id {note_id}:")
                    echo_warn(llm_output_raw)
                else:
                    self.db.upsert_note_feature(note_id, self, llm_output_raw, feature, row.id)
            else:
                echo_warn(f"No LLM output generated for note id {note_id}")


    def run(self, all_note_ids, max_note_ids=None, batch_size=None, dry_run=False):
        if self.within_worker:
            if dry_run: return echo_info("Info: Dry run, exiting without running any LLMs")
            note_ids = all_note_ids
            echo_info(f"Worker started on {gethostname()}")
        else:
            # There is no point running this on notes without the required section
            note_ids = self.db.filter_to_notes_with_section(all_note_ids, self.sec_strat)
            if (num_filtered := len(all_note_ids) - len(note_ids)) > 0:
                echo_warn(f"Skipping {num_filtered} notes without an {self.sec_task_name} section")
            if max_note_ids is not None and (len(note_ids) > max_note_ids):
                echo_warn(f"Limiting to the first {max_note_ids} notes")
                note_ids = note_ids[:max_note_ids]
            if dry_run: return echo_info("Info: Dry run, exiting without running any LLMs")
            
            if self.slurm_enabled:
                # Create a SlurmJobManager and submit jobs that run worker processes
                # To start a worker, you invoke the same command with `--worker` appended
                worker_cmd = sys.argv
                worker_cmd.append("--worker")
                job_manager = SlurmJobManager(worker_cmd, self.config)
                return job_manager.run(note_ids)

        # Either slurm is disabled or this is one of the worker processes.
        # Proceed to running llama.cpp
        self._run_llm(note_ids)


"""LLaCIE: an LLM-based clinical information extraction toolkit.

This package provides numerous modules to support the above functionality
but the primary interface is CLI-based and lives in scripts/llacie.py"""


import os
from dotenv import dotenv_values
from click import BadParameter
from collections import UserDict

from .utils import echo_warn

class ConfigError(BadParameter):
    pass

class Config(UserDict):
    """Holds configuration for the LLaCIE command line app.
    
    Most values are loaded from .env files: example default values are in
    .env.example, and they can be copied to .env and overridden.
    
    This object offers a dictionary-like interface, except that missing
    values return None rather than raising a KeyError.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.data.update({
            **dotenv_values(".env.example"),
            **dotenv_values(".env"),
            **os.environ
        })

        # Set os.environ if we have a HF_TOKEN in a config file, since the `transformers` library
        # will read it from there and not our config files
        if self.data.get('HF_TOKEN') is not None and os.environ.get('HF_TOKEN') is None:
            os.environ['HF_TOKEN'] = self.data.get('HF_TOKEN')

        if self.data.get('PG_URI') is None:
            self.data['PG_URI'] = self._get_pg_connection_string()
        
        if self.data.get('EDW_URI') is None:
            self.data['EDW_URI'] = self._get_edw_connection_string()


    def __missing__(self, key):
        return None
    

    def get(self, key, default=None):
        if key in self.data:
            return self.data[key]
        return default


    def _get_pg_connection_string(self):
        pg_user = self['PG_USER']
        pg_password = self['PG_PASS']
        pg_host = self['PG_HOST']
        pg_dbname = self['PG_DBNAME']

        if pg_host is None or pg_dbname is None:
            raise ConfigError("PG_HOST and PG_DBNAME need to be configured")

        if pg_user is None or pg_password is None or pg_password == '':
            try:
                with open(os.environ['HOME'] + "/.pgpass", "r") as f:
                    for line in f.read().split('\n'):
                        fields = line.strip().split(':')
                        if fields[0] == pg_host and fields[2] == pg_dbname:
                            pg_user = fields[3]
                            pg_password = fields[4]
            except IOError:
                echo_warn("Tried to load postgres login info from ~/.pgpass but it wasn't readable")
        
        if pg_user is None or pg_password is None:
            raise ConfigError("Could not find postgres login information")
        
        return f"postgresql://{pg_user}:{pg_password}@{pg_host}/{pg_dbname}"
    

    def _get_edw_connection_string(self):
        edw_user = self['EDW_USER']
        edw_pass = self['EDW_PASS']
        edw_host = self['EDW_HOST']

        if edw_pass is None:
            try:
                with open(os.environ['HOME'] + "/.edw_env", "r") as f:
                    edw_pass = f.read().split('"')[1]
            except IOError:
                pass

        if edw_user is None or edw_pass is None or edw_host is None:
            return None
        
        return f"mssql+pyodbc://{edw_user}:{edw_pass}@{edw_host}"



import re
from tqdm import tqdm

from ...abstract import AbstractStrategy
from ....tasks.section import ShortHPISectionTask
from ....utils import chunker, echo_info

class ShortHPISectionRegexStrategy(AbstractStrategy):
    """\
    Attempts to extract the `hpi_short` section using regular expressions
    to find most likely beginning and ending delimiters within each note.
    """
    task = ShortHPISectionTask
    name = "section.hpi_short.regex"
    version = "0.0.1"
    BATCH_SIZE = 1000

    prereq_tasks = []

    # Markers for the start of the HPI
    START_REGEX = re.compile(r"""
        \b(
            HPI
            | (History|Central[ ]Elements)[ ]of[ ](the[ ])?(Present(ing)?[ ]Illness|Traumatic[ ]Injury)
            | Brief[ ]summary
        )\b( # Optional prefixes within the HPI text that should be excised
            :
            | \s+Comments:\s*
            | \s+HPI\b
            | \s+History[ ]of[ ](the[ ])?Present[ ]Illness
            | \s+Chief[ ]Complaint:[^\n]*
        )*
        """, re.VERBOSE | re.IGNORECASE)

    # Note that for short HPIs, we will be aggressive about trimming into ED and hospital courses
    # at the end of longer HPIs. Short HPI aims to capture the "meat", often just the first paragraph.
    END_REGEX = re.compile(r"""
        (
            \n ( 
                # Markers for end of HPI when it *begins* a line, w/ exact terminal punctuation
                History[ ]reviewed.[ ]+No[ ]pertinent[ ]past[ ]medical[ ]history.
            )
            | \n (
                # Same, but these terms end with a colon or newline instead of specific punctuation
                # This implies that the entire phrase was "formatted" as a heading or subheading
                Electronic[ ]Medical[ ]Records[ ]Reviewed
                | History[ ]provided[ ]by
                | (E[DWR]|Emergency[ ](Department|Room))[ ]Course
                | Current[ ]Assessment
                | Historical[ ]features
                | Focused[ ]COVID[ ]History
                | (ED[ ]Triage[ ])? (Vitals|Vital[ ]Signs)
                | (Plan|A[/&]P)
                | (Estimated[ ]Date[ ]of[ ]Delivery|EDD)
                | (Relevant|Pertinent[ ])? (Home[ ])? Medications
                | (Relevant|Pertinent[ ])? (Physical[ ])? Exam
                | Quality[ ]Bundle
            ) [:\n]
            | \n (
                # Finally these terms can end on *any word boundary*
                # These words imply high confidence for end of HPI, even when beginning a sentence
                Review[ ]of[ ]Systems
                | ROS
                | Past[ ](Medical[ ])?(History|Hx)
                | Patient[ ]Active[ ]Problem[ ]List
                | Medical (([/\s]+|[ ]+(\band\b|&)[ ]+)Surgical)? [ ](History|Hx)
                | PMH
                | (In[ ]|On[ ]arrival[ ](to|at)[ ]) (the[ ])? (\w+[ ])? 
                    (E[DWR]|Emergency[ ](Room|Department))
                | Data reviewed
                | ((Impression|Assessment)[ ]and[ ]) Plan
            ) \b
        )
        """, re.VERBOSE | re.IGNORECASE)


    @staticmethod
    def clean_note_text(text):
        text = re.sub("  ", "\n", text)
        text = re.sub("\n[ ?]+", "\n", text)
        text = re.sub("\n\n+", "\n\n", text)
        return text.strip()
    

    @classmethod
    def extract_short_hpi(cls, note_text, debug=False):
        start_token = re.search(cls.START_REGEX, note_text)
        if start_token is None: return None
        start_token_end = start_token.end(0)

        end_token = re.search(cls.END_REGEX, note_text[start_token_end:None])
        end_token_start = end_token.start(0) + start_token_end if end_token is not None else None
        if debug: print(start_token, end_token)

        hpi_text = note_text[start_token_end:end_token_start]
        # \xa0 = non-breaking space
        return hpi_text.strip(":?-_ \xa0\n")  


    def run(self, all_note_ids, batch_size=None):
        fail_count = 0
        batch_size = batch_size if batch_size is not None else self.BATCH_SIZE

        echo_info(f"In batches of {batch_size}")
        needs_hpi_pb = tqdm(chunker(all_note_ids, batch_size))

        # Temporarily drop foreign keys during this operation.
        # This speeds up inserts by 10x or more; we don't expect perfect ACID during every task
        with self.db.foreign_keys_temporarily_dropped('note_sections'):
            for note_ids in needs_hpi_pb:
                needs_hpi_pb.set_description(
                    f"Extracting HPIs ({fail_count} failed/{len(all_note_ids)} total)")
                
                df = self.db.get_full_text_notes(note_ids)
                df['note_text'] = [self.clean_note_text(text) for text in df['note_text']]
                df['hpi_short'] = [self.extract_short_hpi(text) for text in df['note_text']]

                for _, row in df.iterrows():
                    if row['hpi_short'] is not None and len(row['hpi_short']) > 0: 
                        self.db.insert_note_section(row['id'], self, row['hpi_short'])
                    else: 
                        fail_count += 1
        
        echo_info(f"Finished! ({fail_count} failed/{len(all_note_ids)} total)")



from abc import ABC, abstractmethod
from copy import copy

class AbstractWorkerCache(ABC):
    """
    The WorkerCache classes implement intermediate data structures used to store inputs and outputs 
    for worker processes during long running processes.
    
    These were invented primarily for the AbstractVllmStrategy pathway of creating features
    using `vllm` on GPU nodes in ERISXdl. These nodes don't have any networking capabilities when
    running, except for the ability to read and write from a shared filesystem. Therefore,
    we can't talk directly to the LlacieDatabase from those nodes, and need an intermediary.
    """
    def __init__(self, upstream_db, config, **options):
        self.upstream_db = upstream_db
        # Caches create a **copy** of the config passed into them to prevent
        # any modifications from propagating upwards (e.g., into the App)
        self.config = copy(config)

    @abstractmethod
    def rename(self): raise NotImplementedError

    @abstractmethod
    def finalize(self): raise NotImplementedError

    @abstractmethod
    def sweep_into_upstream_db(self): raise NotImplementedError


class AbstractFeatureWorkerCache(AbstractWorkerCache):
    """
    These are the methods expected for a WorkerCache that would support generating a feature from
    note sections.
    """

    ### These methods are used by the worker process

    @abstractmethod
    def get_note_section(self): raise NotImplementedError

    @abstractmethod
    def upsert_note_feature(self): raise NotImplementedError

    ### These methods are used by the parent process, which has access to `self.upstream_db`

    @abstractmethod
    def cache_note_sections(self): raise NotImplementedError


"""Integration tests for Quickstart workflow with output and database validation.

This validates each step based on:
1. Command output messages (as shown in app.py)
2. Database table checks where applicable
"""
import subprocess
import pytest
import getpass
import gzip

from pathlib import Path
from contextlib import contextmanager
from sqlalchemy import create_engine, text

# Timeout values, in seconds
DEFAULT_TIMEOUT = 60
LONG_TIMEOUT = 600 

# Expected counts after the various steps of the pipeline are completed
TABLE_COUNT = 9
NOTE_COUNT = 100
SECTION_COUNT = 100
SHORT_FEATURE_COUNT = 2
HUMAN_LABELS_COUNT = 145

# A fixture that contains SQL data to support tests of steps after the LLM runs
SKIPTO_EPISODE_LABELS_SQL = "SKIPTO_episode-labels_extract.sql.gz"


def run_cmd(cmd, timeout=None, capture_output=True):
    if timeout is None: timeout = DEFAULT_TIMEOUT
    kwargs = {"check": True}
    if capture_output: 
        kwargs["check"] = False
        kwargs["text"] = True
    return subprocess.run(cmd, capture_output=capture_output, timeout=timeout, **kwargs)


@contextmanager
def get_db_connection(llacie_config):
    """Get database connection as a context manager.

    Ensures the connection is properly closed when the context exits.
    """
    engine = create_engine(llacie_config["PG_URI"])
    engine.llacie_config = llacie_config
    try:
        yield engine
    finally:
        engine.dispose()


def count_tables(engine):
    """Count the number of tables in the database."""
    query = """
        SELECT COUNT(*)
        FROM information_schema.tables
        WHERE table_schema = 'public'
        AND table_type = 'BASE TABLE'
    """
    with engine.connect() as conn:
        result = conn.execute(text(query))
        return result.scalar()


def count_rows(engine, table_name, where_clause=None):
    """Count rows in a table."""
    prefix = engine.llacie_config["PG_TABLE_PREFIX"]
    query = f"SELECT COUNT(*) FROM {prefix}{table_name}"
    if where_clause:
        query += f" WHERE {where_clause}"
    with engine.connect() as conn:
        result = conn.execute(text(query))
        return result.scalar()


def open_possible_gzip(filepath, mode='rt'):
    """Open file, detecting gzip by magic number."""
    with open(filepath, 'rb') as f:
        magic = f.read(2)
    if magic == b'\x1f\x8b':  # gzip magic number
        return gzip.open(filepath, mode)
    return open(filepath, mode)


def load_sql_file(llacie_config, sql_file):
    """Loads an SQL file (optionally gzipped) into the database."""
    with get_db_connection(llacie_config) as engine:
        conn = engine.raw_connection()
        cursor = conn.cursor()
        with open_possible_gzip(sql_file, 'rt') as f:
            cursor.execute(f.read())
        conn.commit()
        conn.close()


class TestQuickstart:
    """Test the Quickstart workflow with proper validation at each step."""

    @pytest.mark.dependency()
    def test_01_init_db(self, isolated_env, temp_db):
        """Test: llacie init-db

        Expected output: A message for each created table
        Validation: Command succeeds and the expected number of tables exist
        """
        result = run_cmd(["llacie", "init-db"])

        # Validate command return code and outputted messages to stderr
        assert result.returncode == 0, f"init-db failed: {result.stderr}"
        message_count = result.stderr.count("Creating table:")
        assert message_count == TABLE_COUNT, \
            f"Expected {TABLE_COUNT} messages, got {message_count}"

        # Validate database
        with get_db_connection(temp_db) as engine:
            table_count = count_tables(engine)
            assert table_count == TABLE_COUNT, \
                f"Expected {TABLE_COUNT} tables in database, found {table_count}"


    @pytest.mark.dependency()
    def test_02_import_notes(self, isolated_env, temp_db):
        """Test: llacie import-notes text examples/admission-100.txt

        Expected output: "Successfully imported X notes"
        Validation: Check that the notes table has the expected number of rows
        """
        # Setup: init db first
        run_cmd(["llacie", "init-db"])

        notes_path = isolated_env / "examples" / "admission-100.txt"
        result = run_cmd(["llacie", "import-notes", "text", str(notes_path)])

        # Validate command return code
        assert result.returncode == 0, f"import-notes failed: {result.stderr}"

        # Validate output message
        assert "Successfully imported" in result.stderr, \
            f"Expected success message not found in output: {result.stderr}"
        assert f"{NOTE_COUNT} notes" in result.stderr, \
            f"Expected {NOTE_COUNT} notes to be imported, got: {result.stderr}"

        # Validate database
        with get_db_connection(temp_db) as engine:
            note_count = count_rows(engine, "notes")
            assert note_count == NOTE_COUNT, \
                f"Expected {NOTE_COUNT} note rows in database, found {note_count}"

        # Store result
        output_file = isolated_env / "output_02_import_notes.txt"
        output_file.write_text(f"STDOUT:\n{result.stdout}\n\nSTDERR:\n{result.stderr}")


    @pytest.mark.dependency()
    def test_03_sections_extract(self, isolated_env, temp_db):
        """Test: llacie sections extract -s regex

        Expected output: "X notes will undergo section extraction"
        Validation: Check that the note_sections table has the expected number of rows
        """
        # Setup
        run_cmd(["llacie", "init-db"])
        notes_path = isolated_env / "examples" / "admission-100.txt"
        run_cmd(["llacie", "import-notes", "text", str(notes_path)])

        result = run_cmd(["llacie", "sections", "extract", "-s", "regex"])

        # Validate command return code
        assert result.returncode == 0, f"sections extract failed: {result.stderr}"

        # Validate output message
        assert "notes will undergo section extraction" in result.stderr, \
            f"Expected extraction message not found: {result.stderr}"

        # Validate database
        with get_db_connection(temp_db) as engine:
            section_count = count_rows(engine, "note_sections")
            assert section_count == SECTION_COUNT, \
                f"Expected {SECTION_COUNT} note_section rows, found {section_count}"

        # Store result
        output_file = isolated_env / "output_03_sections_extract.txt"
        output_file.write_text(f"STDOUT:\n{result.stdout}\n\nSTDERR:\n{result.stderr}")


    @pytest.mark.dependency(depends=["TestQuickstart::test_03_sections_extract"])
    @pytest.mark.slow
    @pytest.mark.timeout(LONG_TIMEOUT)
    @pytest.mark.requires_llm
    def test_04_features_extract(self, isolated_env, temp_db):
        """Test: llacie features extract -s llama3_8b

        Expected output: "X notes will undergo feature extraction"
        Validation: Check that the note_features table has entries
        """
        # Setup - only run if test_03 hasn't already done this
        # Since we're using class fixtures, we need to ensure the pipeline is run
        run_cmd(["llacie", "init-db"])
        notes_path = isolated_env / "examples" / "admission-100.txt"
        run_cmd(["llacie", "import-notes", "text", str(notes_path)])
        run_cmd(["llacie", "sections", "extract", "-s", "regex"])

        result = run_cmd(
            ["llacie", "features", "extract", "-s", "llama3_8b", "-M", str(SHORT_FEATURE_COUNT)],
            timeout=LONG_TIMEOUT
        )

        # Validate command succeeded
        assert result.returncode == 0, f"features extract failed: {result.stderr}"

        # Validate output message
        assert "notes will undergo feature extraction" in result.stderr, \
            f"Expected extraction message not found: {result.stderr}"

        # Validate database
        with get_db_connection(temp_db) as engine:
            feature_count = count_rows(engine, "note_features")
            assert feature_count > 0, \
                f"Expected >0 note_features in database, found {feature_count}"

        # Store result
        output_file = isolated_env / "output_04_features_extract.txt"
        output_file.write_text(f"STDOUT:\n{result.stdout}\n\nSTDERR:\n{result.stderr}")


    @pytest.mark.dependency()
    def test_05_episode_labels_extract(self, class_isolated_env, class_temp_db):
        """Test: llacie episode-labels extract -s pres_sx_eplab2.llama3_8b

        Expected output: "X episodes will undergo label creation"
        Validation: Check that episode_labels table has entries
        """
        # This test loads a cached copy of the database from after test_04_... completed
        sql_file = Path(__file__).parent.parent / "fixtures" / SKIPTO_EPISODE_LABELS_SQL
        load_sql_file(class_temp_db, sql_file)

        # Run the next step of the llacie pipeline: extracting episode labels
        result = run_cmd(
            ["llacie", "episode-labels", "extract", "-s", "pres_sx_eplab2.llama3_8b"]
        )

        # Validate command succeeded
        assert result.returncode == 0, f"episode-labels extract failed: {result.stderr}"

        # Validate output message (from app.py line 218)
        assert "episodes will undergo label creation" in result.stderr, \
            f"Expected label creation message not found: {result.stderr}"

        # Validate database
        with get_db_connection(class_temp_db) as engine:
            label_count = count_rows(engine, "episode_labels")
            assert label_count > 0, f"Expected >0 episode labels in database, found {label_count}"


    @pytest.mark.dependency(depends=["TestQuickstart::test_05_episode_labels_extract"])
    def test_06_episode_labels_import(self, class_isolated_env, class_temp_db):
        """Test: llacie episode-labels import pres_sx_eplab2 examples/admission-100-labels.xlsx

        Expected output: Import completes successfully
        Validation: Check that human annotations are in database

        NOTE: Uses class-scoped fixtures to reuse database from test_05
        """
        # No setup needed - test_05 already has the full pipeline

        labels_path = class_isolated_env / "examples" / "admission-100-labels.xlsx"
        result = run_cmd(
            ["llacie", "episode-labels", "import", "pres_sx_eplab2", str(labels_path)]
        )

        # Validate command succeeded
        assert result.returncode == 0, f"episode-labels extract failed: {result.stderr}"

        # Validate database
        username = getpass.getuser()
        with get_db_connection(class_temp_db) as engine:
            label_count = count_rows(engine, "episode_labels", f"\"FK_human_annotator\"='{username}'")
            assert label_count == HUMAN_LABELS_COUNT, \
                f"Expected {HUMAN_LABELS_COUNT} human-created episode_labels, found {label_count}"


    @pytest.mark.dependency(depends=["TestQuickstart::test_06_episode_labels_import"])
    def test_07_episode_labels_evaluate(self, class_isolated_env, class_temp_db):
        """Test: llacie episode-labels evaluate

        Expected output: Confusion matrices with accuracy, precision, recall, F1
        Validation: Check that evaluation metrics are present in output

        NOTE: Uses class-scoped fixtures to reuse database from test_06
        """
        # No setup needed - test_06 already has the full pipeline with imported labels!

        result = run_cmd(["llacie", "episode-labels", "evaluate"])

        # Validate command succeeded
        assert result.returncode == 0, f"episode-labels evaluate failed: {result.stderr}"

        # Validate that the output contains evaluation metrics
        output_lower = result.stdout.lower()
        metrics = ["accuracy", "precision", "recall", "f1"]
        found_metrics = [m for m in metrics if m in output_lower]
        assert len(found_metrics) == 4, \
            f"Expected evaluation metrics in output, found only {found_metrics}: {result.stdout}"



import time
from abc import ABC, abstractmethod
from textwrap import dedent
from copy import copy

class AbstractStrategy(ABC):
    def __init__(self, db, config):
        # Strategies are persisted lazily into the database
        # See `get_or_register_task()` and `get_or_register_strategy()` in `LlacieDatabase`
        self.db = db
        # Strategies create a **copy** of the config passed into them to prevent
        # any modifications from propagating upwards (e.g., into the App)
        self.config = copy(config)
        self._start_time = None
        self._within_worker = config['LLACIE_WORKER']

    @property
    @abstractmethod
    def task(self): raise NotImplementedError

    @property
    @abstractmethod
    def name(self): raise NotImplementedError

    @property
    @abstractmethod
    def version(self): raise NotImplementedError

    @property
    @abstractmethod
    def prereq_tasks(self): raise NotImplementedError

    @abstractmethod
    def run(self, ids): raise NotImplementedError

    @property
    def desc(self): return dedent(self.__doc__)

    @property
    def db_row(self): return self.db.get_or_register_strategy(self)

    @property
    def id(self): return self.db_row.id

    @property
    def task_id(self): return self.db.get_or_register_task(self.task).id

    @property
    def within_worker(self): return self._within_worker

    def as_row(self):
        """Fields that get saved to the database"""
        task_row = self.db.get_or_register_task(self.task)
        return {
            "FK_task_id": task_row.id,
            "name": self.name,
            "version": self.version,
            "desc": self.desc
        }

    # TODO: method for checking if the Strategy's prereqs are all complete

    def get_unfinished_ids(self, newer_than=None):
        return self.db.get_unfinished_ids_for_strategy_or_task(self, newer_than=newer_than)
    
    # Methods for timing tasks within runs of a strategy
    # Returns an interval, in seconds, as a float value
    def start_timer(self):
        self._start_time = time.perf_counter()

    def stop_timer(self):
        if self._start_time is None: raise RuntimeError("Strategy run timer wasn't started!")
        runtime = time.perf_counter() - self._start_time
        self._start_time = None
        return runtime


from textwrap import dedent

from ...abstract_llama_cpp import AbstractLlamaCppStrategy
from ...section.hpi_short.regex import ShortHPISectionRegexStrategy
from ....tasks.feature import PresentingSymptomsFeatureTask

class PresentingSxFeatureLLama1Strategy(AbstractLlamaCppStrategy):
    """\
    Attempts to extract the `presenting_sx` feature using LLaMA 1, 13B model,
    with no fine-tuning.

    Uses the `hpi_short` sections created by the section.hpi_short.regex strategy.
    """
    task = PresentingSymptomsFeatureTask
    name = "feature.presenting_sx.llama1"
    version = "0.0.1"
    prereq_tasks = []

    SECTION_STRATEGY_CLASS = ShortHPISectionRegexStrategy
    SLURM_ENABLED = True

    LLAMA_CPP_BIN = 'main-41c6741'
    LLAMA_CPP_THREADS = 16
    SLURM_THREADS_PER_JOB = 16
    LLAMA_MODEL_PATH = 'models/llama/13B/ggml-model-q4_0.bin'
    # Note: this strategy predates the use of grammars to constrain output!
    LLM_MAX_TOKENS = 1020
    LLAMA_CPP_OPTS = f"-m {LLAMA_MODEL_PATH} -n 128 --temp 0.1 --top_p 1.0 -c {LLM_MAX_TOKENS + 4}"
    # The Llama 1 and 2 models specifically use LlamaTokenizer in the transformers module
    LLAMA_VOCAB_PATH = 'models/llama/tokenizer.model'

    LLM_PROMPT = dedent("""\
        {input}

        This patient had the following presenting symptoms: """)



from abc import ABC, abstractmethod
from enum import Enum
from textwrap import dedent

class TaskOutputType(Enum):
    # Constants for the type of task
    # Section tasks divide a note into sections like HPI, Exam, A&P etc.
    SECTION = "section"
    # Feature tasks extract certain pieces of unstructured data from a note, like symptoms
    FEATURE = "feature"
    # Label tasks attempt to condense multiple features per note, patient, or episode
    #   into an overall categorization for the same item
    NOTE_LABEL = "note_label"
    EPISODE_LABEL = "episode_label"
    PATIENT_LABEL = "patient_label"


class AbstractTask(ABC):
    def __init__(self):
        raise RuntimeError('Task classes are not meant to be instantiated')

    @classmethod
    @property
    @abstractmethod
    def output_type(cls): raise NotImplementedError

    @classmethod
    def desc(cls): return dedent(cls.__doc__)

    @classmethod
    def as_row(cls):
        return {
            "output_type": cls.output_type.value,
            "name": cls.name,
            "desc": cls.desc()
        }


class AbstractSectionTask(AbstractTask):
    output_type = TaskOutputType.SECTION

    @classmethod
    @property
    @abstractmethod
    def name(cls): raise NotImplementedError

    @classmethod
    @property
    @abstractmethod
    def note_where_sql(cls): raise NotImplementedError


class AbstractFeatureTask(AbstractTask):
    output_type = TaskOutputType.FEATURE

    @classmethod
    @property
    @abstractmethod
    def name(cls): raise NotImplementedError

    @classmethod
    @property
    @abstractmethod
    def note_join_sql(cls): raise NotImplementedError

    @classmethod
    @property
    @abstractmethod
    def note_where_sql(cls): raise NotImplementedError


class AbstractNoteLabelTask(AbstractTask):
    output_type = TaskOutputType.NOTE_LABEL
    vocab = None
    max_human_labels = 1

    @classmethod
    @property
    @abstractmethod
    def name(cls): raise NotImplementedError

    @classmethod
    @property
    @abstractmethod
    def note_join_sql(cls): raise NotImplementedError

    @classmethod
    @property
    @abstractmethod
    def note_where_sql(cls): raise NotImplementedError


class AbstractEpisodeLabelTask(AbstractTask):
    output_type = TaskOutputType.EPISODE_LABEL
    vocab = None
    max_human_labels = 1

    @classmethod
    @property
    @abstractmethod
    def name(cls): raise NotImplementedError

    @classmethod
    @property
    @abstractmethod
    def episode_join_sql(cls): raise NotImplementedError

    @classmethod
    @property
    @abstractmethod
    def episode_where_sql(cls): raise NotImplementedError

    # Used to find the relevant notes for an episode when creating human annotations
    @classmethod
    @property
    @abstractmethod
    def note_where_sql(cls): raise NotImplementedError


class AbstractPatientLabelTask(AbstractTask):
    output_type = TaskOutputType.PATIENT_LABEL
    vocab = None
    max_human_labels = 1

    @classmethod
    @property
    @abstractmethod
    def name(cls): raise NotImplementedError

    @classmethod
    @property
    @abstractmethod
    def patient_join_sql(cls): raise NotImplementedError

    @classmethod
    @property
    @abstractmethod
    def patient_where_sql(cls): raise NotImplementedError

    # Used to find the relevant notes for a patient when creating human annotations
    @classmethod
    @property
    @abstractmethod
    def note_where_sql(cls): raise NotImplementedError



import os
import tempfile
import pandas as pd
import time

from os import path
from sqlalchemy import create_engine, text
from contextlib import contextmanager
from datetime import datetime
from glob import glob
from collections import defaultdict

from .abstract import AbstractFeatureWorkerCache
from ..batch.slurm import SlurmJobManager
from ..tasks import TaskOutputType as out_t
from ..utils import echo_warn


class SqliteFeatureWorkerCache(AbstractFeatureWorkerCache):
    """
    Implements a WorkerCache for generating features from note sections by saving the inputs
    and outputs in a series of SQLite files (each file is one shard for one worker).
    """

    WORKER_CACHE_SUBDIR = "cache"
    WORKER_CACHE_EXTENSION = "cache.db"
    WORKER_CACHE_DONE_EXTENSION = "done.db"

    WORKER_CACHE_DIRECTORY_RETRIES = 10
    WORKER_CACHE_DIRECTORY_DELAY = 60

    # Change this to True to leave all files on disk, even after being pushed to the upstream DB
    WORKER_CACHE_KEEP_AFTER_SWEEP = False  

    def __init__(self, upstream_db, config, feature_strategy, **options):
        super().__init__(upstream_db, config, **options)

        self._dir = config.get("LLACIE_WORKER_CACHE_DIR", 
            f"{SlurmJobManager.SLURM_IO_DIR}/{self.WORKER_CACHE_SUBDIR}")
        self._ext = config.get("LLACIE_WORKER_CACHE_EXTENSION", self.WORKER_CACHE_EXTENSION)
        self._done_ext = config.get("WORKER_CACHE_DONE_EXTENSION", 
            self.WORKER_CACHE_DONE_EXTENSION)
        self._keep_after_sweep = config.get("LLACIE_WORKER_CACHE_KEEP_AFTER_SWEEP",
            self.WORKER_CACHE_KEEP_AFTER_SWEEP)
        self._retries = int(config.get("LLACIE_WORKER_CACHE_DIRECTORY_RETRIES",
            self.WORKER_CACHE_DIRECTORY_RETRIES))
        self._retry_delay = int(config.get("LLACIE_WORKER_CACHE_DIRECTORY_DELAY",
            self.WORKER_CACHE_DIRECTORY_DELAY))
        
        self.strategy = feature_strategy

        self._name = None
        self._default_shard = None
        self._current_shard = None
        self.prefix = ""  # For table names; they're less important here than in LlacieDatabase

        # This tracks the status of shards. Note that it is lazy and does not necessarily reflect
        # changes that other processes may have made to the filesystem. These can only be detected
        # and updated into this variable using the `self.find_finalized_shards()` generator.
        self._is_done = defaultdict(lambda: False)

        # If LLACIE_WORKER_CACHE_PATH is specified in the environment or the config at startup,
        # the cache directory, cache name, and current shard are parsed from this path.
        # This is how individual workers get automatically pointed to the right shard.
        prespecified_path = config.get("LLACIE_WORKER_CACHE_PATH")
        if prespecified_path is not None: 
            self._name, shard, is_done = self._parse_path(prespecified_path)
            if is_done is None: 
                raise RuntimeError("An invalid LLACIE_WORKER_CACHE_PATH was provided")
            self._current_shard = self._default_shard = shard
            self._is_done[shard] = is_done

        if not self._autocreate_directory():
            raise RuntimeError("Could not autocreate the directory for the Llacie worker cache")


    def _parse_path(self, shard_path):
        """
        Parse a given `shard_path` to a shard file in the cache into the corresponding cache name, 
        shard, and shard status.

        If there is a problem with parsing, this will return None for all components.
        """
        filename_parts = path.basename(shard_path).split(".", 1)
        shard = filename_parts[0]
        if not shard.isdigit(): return (None, None, None)
        name = path.basename(path.dirname(shard_path))
        is_done = filename_parts[1] == self._done_ext
        if not is_done and filename_parts[1] != self._ext: return (None, None, None)
        return (name, shard, is_done)
    

    def _check_if_done(self, shard):
        is_done = path.exists(self.get_path(for_shard=shard, is_done=True))
        self._is_done[shard] = is_done
        return is_done


    def _autocreate_directory(self):
        """
        Ensure that the directories required by the cache exist.
        
        If the cache does not yet have a name (they are typically named by Slurm job ID, but this
        may not be available before the job is submitted), a random one is created using mkdtemp.
        The name can be updated later with `self.rename()`.
        """
        os.makedirs(self._dir, exist_ok = True)
        if self._name is None:
            self._name = path.basename(tempfile.mkdtemp(prefix="unsubmitted-", dir=self._dir))
            os.chmod(path.join(self._dir, self._name), 0o700)
            return True

        dir_path = path.join(self._dir, self._name)
        # Workers can step on each other in accessing this directory for the first time (on NFS)
        # Therefore we allow a few retries
        for attempt in range(self._retries):
            try:
                os.makedirs(dir_path, exist_ok = True)
                return True
            except PermissionError:
                echo_warn(f"Permissions error in opening worker cache directory {dir_path}.")
                if attempt == self._retries - 1: return False
                echo_warn(f"Attempt {attempt + 1}, retrying in {self._retry_delay} seconds...")
                time.sleep(self._retry_delay)
        
        return False

    @contextmanager
    def conn(self, shard=None):
        """
        Provides a connection to a shard, by default the current shard. The current shard can be 
        changed by providing a `shard` argument, which can be any number or numeric string, but 
        typically these are sequential integers corresponding to Slurm job array task IDs.
        
        Intended to be used as a context manager, e.g., `with self.conn() as conn:` to ensure that
        the shard's corresponding SQLite database file is closed after any operations.
        
        Note that `conn.commit()` MUST be called by the user after any query that makes changes
        to the database, otherwise the changes will be rolled back!
        """
        if shard is not None:
            if not str(shard).isdigit(): raise RuntimeError("shard must be a numeric string")
            self._current_shard = str(shard)
        if self._current_shard is None: self._current_shard = self._default_shard
        if self._current_shard is None: raise RuntimeError("Unable to select a WorkerCache shard")
        mode = "ro" if self._check_if_done(self._current_shard) else "rwc"
        uri = f"sqlite:///file:{self.get_path()}?mode={mode}&uri=true"
        conn = None

        try:
            conn = create_engine(uri).connect()
            self._autocreate_tables(conn)
            yield conn
        finally:
            if conn is not None: conn.close()


    def get_path(self, for_name=None, for_shard=None, is_done=None):
        """Returns a full path to the current cache file being used, OR for the specific cache
        name, shard, and shard status given, when those are passed in as the optional arguments.
        If not given, these default to whatever the current values are."""
        if for_name is None: for_name = self._name
        if for_shard is None: for_shard = self._current_shard
        if is_done is None: is_done = self._is_done[for_shard]
        ext = self._done_ext if is_done else self._ext
        return os.path.join(self._dir, for_name, f"{for_shard}.{ext}")
    

    def _autocreate_tables(self, conn):
        create_table_sqls = [
            text(f"""\
                CREATE TABLE IF NOT EXISTS "{self.prefix}strategies" (
                    "id" BIGINT UNIQUE NOT NULL,
                    "FK_task_id" BIGINT,
                    "name" VARCHAR(255) NOT NULL,
                    "desc" TEXT,
                    "version" VARCHAR(255) NOT NULL,
                    "last_updated" DATETIME NOT NULL
                );"""),
            text(f"""\
                CREATE TABLE IF NOT EXISTS "{self.prefix}note_sections" (
                    "id" BIGINT UNIQUE NOT NULL,
                    "FK_note_id" BIGINT,
                    "section_name" VARCHAR(40),
                    "section_value" TEXT
                );"""),
            text(f"""\
                CREATE TABLE IF NOT EXISTS "{self.prefix}note_features" (
                    "FK_note_id" BIGINT NOT NULL,
                    "feature_name" VARCHAR(255) NOT NULL,
                    "FK_note_section_id" BIGINT,
                    "FK_strategy_id" BIGINT NOT NULL,
                    "llm_output_raw" TEXT NOT NULL,
                    "feature_value" TEXT,
                    "feature_updated" DATETIME NOT NULL,
                    "strategy_runtime" DECIMAL(12,5),
                    UNIQUE ("FK_note_id", "feature_name", "FK_strategy_id"),
                    FOREIGN KEY ("FK_strategy_id") REFERENCES "{self.prefix}strategies"("id"),
                    FOREIGN KEY ("FK_note_section_id") REFERENCES "{self.prefix}note_sections"("id")
                );""")
        ]
        for sql in create_table_sqls: conn.execute(sql)
        conn.commit()


    def rename(self, new_name):
        """
        Renames the entire cache. Intended to be used when the Slurm array job ID is known and
        you want to move the cache to a path that includes this array job ID so that the workers
        can find it.
        """
        new_name = str(new_name)
        old_dir = os.path.join(self._dir, self._name)
        os.rename(old_dir, os.path.join(self._dir, new_name))
        self._name = new_name
    

    def find_finalized_shards(self):
        """
        A generator that iterates through all of the shards that are currently finalized but not
        yet cleared from the filesystem.
        """
        finalized_shard_files = glob(self.get_path(for_shard='*', is_done=True))
        for shard_file in finalized_shard_files:
            _, shard, _ = self._parse_path(shard_file)
            self._is_done[shard] = True
            yield shard
    

    def _delete_shard(self, shard):
        os.unlink(self.get_path(for_shard=shard))


    ##############
    # These methods are used by the worker process
    ##############

    def get_note_section(self, note_id, strategy):
        if strategy.task.output_type != out_t.SECTION:
            raise RuntimeError("Only strategies for creating sections are allowed.")

        params = {
            "note_id": int(note_id), 
            "section_name": strategy.task.name
        }
        select_section_sql = text(f"""
            SELECT "id", "FK_note_id", "section_name", "section_value"
                FROM "{self.prefix}note_sections"
                WHERE "FK_note_id" = :note_id
                    AND "section_name" = :section_name
                ORDER BY "id" ASC
            """)

        with self.conn() as conn:
            return conn.execute(select_section_sql, params).fetchone()


    def upsert_note_feature(self, note_id, strategy, llm_output_raw, feature_value=None, 
            note_section_id=None, runtime=None):
        if strategy.name != self.strategy.name:
            raise RuntimeError("Inconsistent strategy supplied to upsert_note_feature!")
        if runtime is None:
            try: runtime = strategy.stop_timer()
            except RuntimeError: runtime = None

        get_strategy_params = {"strategy_name": self.strategy.name}
        get_strategy_sql = text(f"""
            SELECT "id", "name" from "{self.prefix}strategies"
                WHERE "name" = :strategy_name
            """)
        update_params = {
            "note_id": int(note_id), 
            "feature_name": strategy.task.name,
            "note_section_id": int(note_section_id),
            "llm_output_raw": llm_output_raw, 
            "feature_value": feature_value,
            "start_time": datetime.now(),
            "strategy_runtime": runtime
        }
        update_nf_sql = text(f"""
            INSERT INTO "{self.prefix}note_features" ("FK_note_id", "feature_name",
                    "FK_strategy_id", "FK_note_section_id", "llm_output_raw", "feature_value",
                    "feature_updated", "strategy_runtime")
                VALUES (:note_id, :feature_name, :strategy_id, :note_section_id, 
                    :llm_output_raw, :feature_value, :start_time, :strategy_runtime)
                ON CONFLICT ("FK_note_id", "feature_name", "FK_strategy_id")
                DO UPDATE SET "FK_note_section_id" = :note_section_id, 
                    "llm_output_raw" = :llm_output_raw,
                    "feature_value" = :feature_value,
                    "feature_updated" = :start_time,
                    "strategy_runtime" = :strategy_runtime
            """)
        
        with self.conn() as conn:
            strat_row = conn.execute(get_strategy_sql, get_strategy_params).fetchone()
            if strat_row is None:
                raise RuntimeError("Could not retrieve the strategy from the WorkerCache!")
            update_params["strategy_id"] = strat_row.id
            conn.execute(update_nf_sql, update_params)
            conn.commit()
    

    def finalize(self): 
        """
        Takes the current shard and marks it as done.

        In practice, this renames the file to end with ".done.db" which signals to other
        processes that it won't be written to any more (and is ready to be saved upstream). From
        this point onward, that shard can only be opened in read-only mode.
        """
        if self._is_done[self._current_shard] is True: 
            raise RuntimeError("finalize() can be called only once per shard!")
        os.rename(self.get_path(), self.get_path(is_done=True))
        self._is_done[self._current_shard] = True


    ##############
    # These methods are used by the parent process, which does have access to `self.upstream_db`
    ##############

    def cache_note_sections(self, id_chunks):
        """
        Saves all of the note sections specified for the note IDs in `id_chunks` into separate
        shards within this cache.
        
        We also cache the single row for the `feature_strategy` that was given to this WorkerCache.
        This is because its database row ID is also needed by workers, to preserve foreign keys.
        """
        strat_df = pd.DataFrame([dict(self.strategy.db_row._mapping)])
        for i, note_ids in enumerate(id_chunks):
            shard = i + 1
            sections_df = self.upstream_db.get_note_sections(note_ids, self.strategy.sec_strat)
            with self.conn(shard) as conn:
                strat_df.to_sql(f"{self.prefix}strategies", conn, if_exists='replace')
                sections_df.to_sql(f"{self.prefix}note_sections", conn, if_exists='replace')
                conn.commit()
    

    def sweep_into_upstream_db(self):
        count = 0
        for shard in self.find_finalized_shards():
            with self.conn(shard) as conn:
                # Dump all the completed note features into a DataFrame
                get_nf_sql = text(f"""
                    SELECT "FK_note_id", "llm_output_raw", "feature_value", "FK_note_section_id",
                        "strategy_runtime"
                        FROM "{self.prefix}note_features"
                    """)
                df = pd.read_sql_query(get_nf_sql, conn)
                # Push the contents of this DataFrame to the upstream database
                for i, row in df.iterrows():
                    self.upstream_db.upsert_note_feature(row['FK_note_id'], self.strategy,
                        row['llm_output_raw'], row['feature_value'], row['FK_note_section_id'],
                        row['strategy_runtime'])
            if not self._keep_after_sweep:
                self._delete_shard(shard)
            count += 1
        return count



import re
import pickle
import platformdirs
import pandas as pd

from os import path, makedirs
from collections import defaultdict
from copy import deepcopy

from .utils import echo_info, echo_warn, PACKAGE_DIR


class Vocab:
    """Implements a vocabulary containing terms, used for labeling tasks. The tabular
    form of each vocabulary is stored in XLSX files under vocabs/. See those files for
    examples of how to format them.
    
    This class uses a pickle file (cache) to avoid reparsing the XLSX whenever possible."""

    def __init__(self, from_file, sheet_name=0, **options):
        self._terms = defaultdict(set)
        self._ngram_dicts = []

        self.from_file = path.join(PACKAGE_DIR, "vocabs", from_file)
        cache_dir = platformdirs.user_cache_dir("llacie", "tpaklab")
        self.cache_file = path.join(cache_dir, "vocabs", f"{from_file}.pkl")

        if self._load_from_cache():
            return
        elif self.from_file.endswith(".xlsx") or self.from_file.endswith(".xls"):
            echo_info(f"Parsing and caching vocabulary in vocabs/{from_file}")
            self._parse_df(pd.read_excel(self.from_file, sheet_name=sheet_name))
        else:
            raise NotImplementedError

    def __contains__(self, term):
        return term in self._terms

    def __len__(self):
        return len(self._terms)
    

    @property
    def terms(self):
        """Returns a list of all (canonical) terms in this vocab, omitting synonyms."""
        return sorted(self._terms.keys())
    
    @property
    def terms_and_synonyms(self):
        """Returns a dict of terms mapped to corresponding lists of synonyms. This is a copy,
        so modifying it has no effect on this Vocab instance."""
        return deepcopy(self._terms)


    def _load_from_cache(self):
        if not path.isfile(self.cache_file): return False
        if path.getsize(self.cache_file) == 0: return False
        if path.getmtime(self.cache_file) < path.getmtime(self.from_file): return False
        try:
            with open(self.cache_file, 'rb') as f:
                loaded_obj = pickle.load(f)
                self._terms = loaded_obj["_terms"]
                self._ngram_dicts = loaded_obj["_ngram_dicts"]
        except (pickle.UnpicklingError, PermissionError, TypeError, KeyError) as e:
            echo_warn(f"Error loading cached vocab: {e}")
            return False
        return True


    def _save_to_cache(self):
        makedirs(path.dirname(self.cache_file), exist_ok=True)
        with open(self.cache_file, 'wb') as f:
            pickled = {"_terms": self._terms, "_ngram_dicts": self._ngram_dicts}
            pickle.dump(pickled, f)


    def _add_terms(self, terms, synonyms=None):
        if synonyms is not None:
            if not isinstance(synonyms, list):
                synonyms = [synonyms]
        for term in terms:
            self._terms[term].add(term)
            self._terms[term].update(synonyms)


    def _parse_df(self, vocab_df):
        max_n = max(vocab_df['n'])
        for n in range(max_n, 0, -1):
            ngram_dict = {}
            for _, row in vocab_df[vocab_df.n == n].iterrows():
                terms = [row['canonical_name']]
                if pd.notna(row['combo_symptom']):
                    terms.append(row['combo_symptom'])
                self._add_terms(terms, row['ngram'])
                ngram_dict[tuple(row['ngram'].split(' '))] = terms
            self._ngram_dicts.append(ngram_dict)
        self._save_to_cache()


    def find_terms_in_feature(self, feature_value):
        """Given a textual feature, searches each line of the feature for terms from this
        vocabulary, preferring longer n-gram matches over shorter ones. This is akin to the
        backoff techniques used in NLP: https://www.scaler.com/topics/nlp/backoff-in-nlp/
        
        Returns a dict() with matched terms as keys and earliest line # matched as values."""

        lines = feature_value.split("\n")
        found_terms = {}
        for line_no in range(len(lines) - 1, -1, -1):
            tokens = re.split(r'\s+', re.sub(r'[^a-z0-9]+', ' ', lines[line_no].lower()).strip())
            while len(tokens) > 0:
                for ngram_i, ngram_dict in enumerate(self._ngram_dicts):
                    n = len(self._ngram_dicts) - ngram_i
                    vocab_matches = ngram_dict.get(tuple(tokens[0:n]), None)
                    if vocab_matches is None: continue
                    for match in vocab_matches:
                        found_terms[match] = line_no + 1
                    tokens = tokens[n:]
                    break
                if vocab_matches is None: tokens.pop(0)
        return found_terms


from textwrap import dedent

from ...abstract_vllm_or_lcp import AbstractVllmOrLcpStrategy
from ...section.hpi_short.regex import ShortHPISectionRegexStrategy
from ....tasks.feature import PresentingSymptomsFeatureTask

class PresentingSxFeatureLlama3Instruct8BStrategy(AbstractVllmOrLcpStrategy):
    """\
    Attempts to extract the `presenting_sx` feature using Llama 3 Instruct 8B model,
    an 8B parameter model in the Llama family released by Meta in April 2024. 

    This uses the original weights casted to torch.float16, because the Tesla V100s that we can 
    run this on in ERISXdl do not support bfloat16. 
    https://huggingface.co/meta-llama/Meta-Llama-3-8B-Instruct

    Also uses the `hpi_short` sections created by the section.hpi_short.regex strategy.
    """
    task = PresentingSymptomsFeatureTask
    name = "feature.presenting_sx.llama3_8b"
    version = "0.0.1"
    prereq_tasks = []

    SECTION_STRATEGY_CLASS = ShortHPISectionRegexStrategy

    # HuggingFace respository ID for the full PyTorch/Transformers model, which vllm uses
    MODEL_ID = 'meta-llama/Meta-Llama-3-8B-Instruct'
    # As a fallback, you can also provide a GGUF version, which is what llama-cpp-python uses
    GGUF_MODEL_ID = 'QuantFactory/Meta-Llama-3-8B-Instruct-GGUF'
    GGUF_MODEL_FILENAME = '*.Q5_K_S.gguf'
    
    SLURM_THREADS_PER_JOB = 16
    ARRAY_OF_TEN_SHORT_STRINGS_JSON_SCHEMA = """
    {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "array",
        "minItems": 1,
        "maxItems": 10,
        "items": {
            "type": "string",
            "minLength": 2,
            "maxLength": 100
        }
    }
    """
    JSON_SCHEMA = ARRAY_OF_TEN_SHORT_STRINGS_JSON_SCHEMA

    VLLM_ENABLED = False
    VLLM_DTYPE = 'half'
    LLM_TOKENS_TO_GENERATE = 1024
    # Arguments to SamplingParams in vllm
    # See: https://docs.vllm.ai/en/latest/api/vllm/#vllm.SamplingParams
    # These are translated into equivalent arguments for llama-cpp-python's create_chat_completion
    SAMPLING_PARAMS = {
        "temperature": 0.1,
        "top_p": 0.8,
        "repetition_penalty": 1.05
    }

    NICE_MODEL_NAME = "Llama3 8B Instruct"

    # Model context length. Note that Vllm will not generate past the end of the context window.
    LLM_MAX_MODEL_LENGTH = 3044  
    LLM_SYSTEM_PROMPT = dedent("""\
        You are a clinical researcher that reads medical charts and answers questions about them. 
        Use only the information in the text provided to answer the question.
        If a patient denies something, do not include it in your answer.
        After you provide an answer, you immediately stop talking.""")
    LLM_USER_PROMPT = dedent("""\
        Read the following patient history and list the patient's presenting symptoms.
        Include only symptoms present now or reported for the days to weeks leading up to admission.
        Ignore any symptoms from past medical history or prior hospital admissions.
        Give your answer as a JSON array containing up to ten strings.
        Each string contains between one and three words.
        
        {input}""")



import os
import re

from transformers import AutoTokenizer
from json.decoder import JSONDecodeError

##### Note that ordinarily we should place all imports at the top, per PEP8
##### But in this unusual case, we avoid importing the `vllm` and `outlines` packages up here
##### because they ONLY exist inside the containerized environment used by worker processes

from ..utils import echo_err

class Vllm(object):
    """
    Runs LLMs on the ERISXdl GPU cluster using the `vllm` package.
    This package is one of the fastest LLM inference packages, with loads of optimizations
        for high throughput and running slimmed-down CUDA kernels on NVIDIA GPUs
    Structured outputs are also supported using the `outlines` package and JSON schemas.
    See: https://docs.vllm.ai/en/latest/index.html
    """
    
    HOME_DIR = os.path.expanduser('~')
    ARRAY_OF_STRINGS_JSON_SCHEMA = """
    {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "array",
        "items": {
            "type": "string"
        }
    }
    """
    JSON_SCHEMA = ARRAY_OF_STRINGS_JSON_SCHEMA


    def __init__(self, config, model_path, system_prompt, user_prompt, json_schema=None, 
            **engine_args):
        """
        Instantiates an instance of the engine. For a comprehensive list of `engine_args` see:
        https://docs.vllm.ai/en/stable/models/engine_args.html
        """
        self.config = config
        self.model_path = model_path
        self.system_prompt = system_prompt
        self.user_prompt = user_prompt
        self.json_schema = json_schema if json_schema is not None else self.JSON_SCHEMA

        self.tokenizer = AutoTokenizer.from_pretrained(
            model_path, 
            local_files_only = model_path.count('/') > 1
        )

        from vllm import LLM
        self._model = LLM(model=model_path, **engine_args)

    
    def create_prompt(self, **kwargs):
        messages = [
            {"role": "system", "content": self.system_prompt.format(**kwargs)},
            {"role": "user", "content": self.user_prompt.format(**kwargs)}
        ]
        return self.tokenizer.apply_chat_template(
            messages,
            tokenize = False, 
            add_generation_prompt = True
        )


    def autotrim_prompt(self, prompt_max_tokens, trim_key="input", **kwargs):
        """Creates a prompt and ensures it is within the token limit.

        `kwargs` are substituted into the chat template created by `system_prompt` and
        by `user_prompt`, using Python's native str.format().

        `prompt_max_tokens` is the token limit, usually $context_window - 4.

        If the generated prompt is longer than `prompt_max_tokens` after tokenization, the item 
        in the dict() `kwargs` at key `trim_key` is trimmed by about one sentence at a time until 
        the threshold is reached. By default, `kwargs["input"]` is trimmed.
        
        Returns the full prompt with the trimmed inputs inserted, or None if the process fails."""
        trimmed_text = kwargs.get(trim_key, "")
        if self.tokenizer is None:
            echo_err("A tokenizer for this model could not be loaded.")
            return None
        while True:
            full_prompt = self.create_prompt(**kwargs)
            if len(self.tokenizer.encode(full_prompt)) <= prompt_max_tokens: break
            # If it's too long, cleave off the last sentence-like piece, and try again
            sentences_and_periods = re.split(r'([.]\s+|[.]$)', trimmed_text)
            if len(sentences_and_periods) < 4:
                return None
            else: 
                trimmed_text = ''.join(sentences_and_periods[:-4]) + '.'
                kwargs[trim_key] = trimmed_text
        return full_prompt
    

    def __call__(self, prompts, **sampling_params):
        from vllm import SamplingParams
        from vllm.sampling_params import StructuredOutputsParams
        struct_out_params = StructuredOutputsParams(json=self.json_schema)
        sampling_params = SamplingParams(**sampling_params, structured_outputs=struct_out_params)

        try:
            outputs = self._model.generate(prompts, sampling_params=sampling_params)
        except JSONDecodeError as e:
            echo_err(f"Invalid JSON. '{e.msg}' at char {e.pos}:\n{e.doc}")
            return [None] * len(prompts)

        return outputs


from .abstract import AbstractEpisodeLabelTask

from ..vocab import Vocab

class PresentingSymptomsEpisodeLabelV1Task(AbstractEpisodeLabelTask):
    """\
    Presenting symptoms for the given episode, limiting to patients with suspected infection.
    
    This implementation uses the original vocab generated for the K08 submission. This vocab
    is expected to be updated periodically, which will bump the number at the end of this class 
    name and the `name` attribute below.

    Typically, this task is implemented by examining features generated for the H&P notes
    (e.g., `PresentingSymptomsFeatureTask`).
    """
    name = "pres_sx_eplab1"
    max_human_labels = 10
    episode_join_sql = """
        LEFT JOIN "{{cohort_table}}" AS esc ON
            e."id" = esc."FK_episode_id"
    """
    episode_where_sql = """
        "infectionCriteria" IS TRUE
    """
    note_where_sql = """
        "InpatientNoteTypeDSC" IN ('H&P')
        AND EXTRACT(EPOCH FROM ("DateOfServiceDTS" - "startDTS")) < 24 * 60 * 60
    """

    vocab = Vocab("dt.pres_sx.ngrams.v1.xlsx", sheet_name="symptom_ngrams_top75ile")


class PresentingSymptomsEpisodeLabelV2Task(PresentingSymptomsEpisodeLabelV1Task):
    """\
    Presenting symptoms for the given episode, limiting to patients with suspected infection.
    
    This implementation uses an **updated vocab**, trying to simplify some of the ambiguities in
    the earlier vocab via additional curation (mostly merging and trimming terms).
    """
    name = "pres_sx_eplab2"

    vocab = Vocab("dt.pres_sx.ngrams.v2.xlsx", sheet_name="symptom_ngrams_top75ile.edited")


class PresentingSymptomsEpisodeLabelV2Top30Task(PresentingSymptomsEpisodeLabelV2Task):
    """\
    Presenting symptoms for the given episode, limiting to patients with suspected infection, and
    using the V2 updated vocab but also restricting to the top 30 most common terms, which we 
    pulled out for further analysis/illustration in the JAMA NO manuscript. These terms are 
    explicitly whitelisted in the attribute below.
    """
    name = "pres_sx_eplab2_top30"

    term_whitelist = {
        "fever", "dyspnea", "cough", "abdominal pain", "pain", "nausea", "chills", "fatigue",
        "vomiting", "altered mental status", "diarrhea", "chest pain", "weakness", "swelling",
        "sputum changes", "redness", "headache", "poor appetite", "malaise", "back pain",
        "hypoxemia", "hypotension", "dizziness", "dysuria", "rhinorrhea", "lightheadedness",
        "fall", "urinary frequency", "diaphoresis", "leg swelling"        
    }